@@ -1,6 +1,6 @@
 /*
  * cups - Python bindings for CUPS
- * Copyright (C) 2002, 2005, 2006, 2007  Tim Waugh <twaugh@redhat.com>
+ * Copyright (C) 2002, 2005, 2006, 2007, 2008, 2011  Tim Waugh <twaugh@redhat.com>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -259,6 +259,62 @@ PPD_localize (PPD *self)
   return PyErr_SetFromErrno (PyExc_RuntimeError);
 }
 
+static PyObject *
+PPD_localizeIPPReason (PPD *self, PyObject *args, PyObject *kwds)
+{
+  PyObject *ret;
+  const char *reason;
+  const char *scheme = NULL;
+  char *buffer;
+  const size_t bufsize = 1024;
+  static char *kwlist[] = { "reason", "scheme", NULL };
+
+  if (!PyArg_ParseTupleAndKeywords (args, kwds, "s|s", kwlist,
+				    &reason, &scheme))
+    return NULL;
+
+  buffer = malloc (bufsize);
+  if (ppdLocalizeIPPReason (self->ppd, reason, scheme, buffer, bufsize))
+  {
+    ret = make_PyUnicode_from_ppd_string (self, buffer);
+  } else {
+    ret = Py_None;
+    Py_INCREF (Py_None);
+  }
+
+  free (buffer);
+  return ret;
+}
+
+static PyObject *
+PPD_localizeMarkerName (PPD *self, PyObject *args)
+{
+#if CUPS_VERSION_MAJOR > 1 || (CUPS_VERSION_MAJOR == 1 && CUPS_VERSION_MINOR >= 4)
+  PyObject *ret;
+  const char *name;
+  const char *lname;
+
+  if (!PyArg_ParseTuple (args, "s", &name))
+    return NULL;
+
+  lname = ppdLocalizeMarkerName (self->ppd, name);
+
+  if (lname != NULL)
+  {
+    ret = make_PyUnicode_from_ppd_string (self, lname);
+  } else {
+    ret = Py_None;
+    Py_INCREF (Py_None);
+  }
+
+  return ret;
+#else /* earlier than CUPS 1.4 */
+  PyErr_SetString (PyExc_RuntimeError,
+		   "Operation not supported - recompile against CUPS 1.4 or later");
+  return NULL;
+#endif /* CUPS 1.4 */
+}
+
 static PyObject *
 PPD_markDefaults (PPD *self)
 {
@@ -451,6 +507,169 @@ PPD_nondefaultsMarked (PPD *self)
   return PyBool_FromLong (nondefaults_marked);
 }
 
+#if !(_POSIX_C_SOURCE >= 200809L || _XOPEN_SOURCE >= 700)
+/*
+ * A rudimentary emulation of getline() for systems that dont support it
+ * natively.  Since this is used for PPD file reading, it assumes (possibly
+ * falsely) that BUFSIZ is big enough.
+ */
+ssize_t
+getline(char **line, size_t *linelen, FILE *fp)
+{
+  if (*linelen == 0) {
+    *linelen = BUFSIZ;
+    *line = malloc(*linelen);
+  }
+
+  memset(*line, 0, *linelen);
+  fgets(*line, *linelen, fp);
+
+  return (strlen(*line));
+}
+#endif
+
+/*
+ * emit marked options by returning a string.
+ */
+
+static PyObject *
+PPD_emitString (PPD *self, PyObject *args)
+{
+  ppd_section_t section;
+  float min_order;
+  char *emitted;
+  PyObject *ret;
+
+  if (!PyArg_ParseTuple (args, "if", &section, &min_order))
+    return NULL;
+
+  emitted = ppdEmitString(self->ppd, section, min_order);
+
+  if (emitted) {
+    ret = PyString_FromString(emitted);
+    free (emitted);
+  } else {
+    ret = Py_None;
+  }
+
+  Py_INCREF (ret);
+
+  return ret;
+}
+
+/*
+ * emit marked options by writing to a file object.
+ */
+
+static PyObject *
+PPD_emit (PPD *self, PyObject *args)
+{
+  PyObject *pyFile;
+  ppd_section_t section;
+  FILE *f;
+
+  if (!PyArg_ParseTuple (args, "Oi", &pyFile, &section))
+    return NULL;
+
+  f = PyFile_AsFile(pyFile);
+  if (!f)
+    return NULL;
+
+  if (!ppdEmit(self->ppd, f, section))
+    return Py_None;
+  return PyErr_SetFromErrno (PyExc_RuntimeError);
+}
+
+/*
+ * emit marked options after order dependency by writing to a file object.
+ */
+
+static PyObject *
+PPD_emitAfterOrder (PPD *self, PyObject *args)
+{
+  PyObject *pyFile;
+  ppd_section_t section;
+  FILE *f;
+  int limit;
+  float min_order;
+
+  if (!PyArg_ParseTuple (args, "Oiif", &pyFile, &section, &limit, &min_order))
+    return NULL;
+
+  f = PyFile_AsFile(pyFile);
+  if (!f)
+    return NULL;
+
+  if (!ppdEmitAfterOrder(self->ppd, f, section, limit, min_order))
+    return Py_None;
+  return PyErr_SetFromErrno (PyExc_RuntimeError);
+}
+
+/*
+ * emit marked options by writing to a file descriptor.
+ */
+
+static PyObject *
+PPD_emitFd (PPD *self, PyObject *args)
+{
+  ppd_section_t section;
+  int f;
+
+  if (!PyArg_ParseTuple (args, "ii", &f, &section))
+    return NULL;
+
+  if (!ppdEmitFd(self->ppd, f, section))
+    return Py_None;
+  return PyErr_SetFromErrno (PyExc_RuntimeError);
+}
+
+/*
+ * emit JCL options by writing to a file object.
+ */
+
+static PyObject *
+PPD_emitJCL (PPD *self, PyObject *args)
+{
+  PyObject *pyFile;
+  int job_id;
+  const char *user;
+  const char *title;
+  FILE *f;
+
+  if (!PyArg_ParseTuple (args, "Oiss", &pyFile, &job_id, &user, &title))
+    return NULL;
+
+  f = PyFile_AsFile(pyFile);
+  if (!f)
+    return NULL;
+
+  if (!ppdEmitJCL(self->ppd, f, job_id, user, title))
+    return Py_None;
+  return PyErr_SetFromErrno (PyExc_RuntimeError);
+}
+
+/*
+ * emit JCL end by writing to a file object.
+ */
+
+static PyObject *
+PPD_emitJCLEnd (PPD *self, PyObject *args)
+{
+  PyObject *pyFile;
+  FILE *f;
+
+  if (!PyArg_ParseTuple (args, "O", &pyFile))
+    return NULL;
+
+  f = PyFile_AsFile(pyFile);
+  if (!f)
+    return NULL;
+
+  if (!ppdEmitJCLEnd(self->ppd, f))
+    return Py_None;
+  return PyErr_SetFromErrno (PyExc_RuntimeError);
+}
+
 PyObject *
 PPD_writeFd (PPD *self, PyObject *args)
 {
@@ -463,7 +682,7 @@ PPD_writeFd (PPD *self, PyObject *args)
     return NULL;
 
   dfd = dup (fd);
-  if (!dfd) {
+  if (dfd == -1) {
     PyErr_SetFromErrno (PyExc_RuntimeError);
     return NULL;
   }
@@ -489,7 +708,8 @@ PPD_writeFd (PPD *self, PyObject *args)
       for (end = start; *end; end++)
 	if (isspace (*end) || *end == ':')
 	  break;
-      keyword = strndup (start, end-start);
+      keyword = calloc(1, (end - start) + 1);
+      strncpy(keyword, start, end - start);
       choice = ppdFindMarkedChoice (self->ppd, keyword);
 
       // Treat PageRegion, PaperDimension and ImageableArea specially:
@@ -619,6 +839,16 @@ PyMethodDef PPD_methods[] =
       "localize() -> None\n\n"
       "Localize PPD to the current locale." },
 
+    { "localizeIPPReason",
+      (PyCFunction) PPD_localizeIPPReason, METH_VARARGS | METH_KEYWORDS,
+      "localizeIPPReason(reason, scheme) -> string or None\n\n"
+      "Localize IPP reason to the current locale." },
+
+    { "localizeMarkerName",
+      (PyCFunction) PPD_localizeMarkerName, METH_VARARGS,
+      "localizeMarkerName(name) -> string or None\n\n"
+      "Localize marker name to the current locale." },
+
     { "markDefaults",
       (PyCFunction) PPD_markDefaults, METH_NOARGS,
       "markDefaults() -> None\n\n"
@@ -672,6 +902,67 @@ PyMethodDef PPD_methods[] =
       "nondefaultsMarked() -> boolean\n\n"
       "Returns true if any non-default option choices are marked." },
 
+    { "emitString",
+      (PyCFunction) PPD_emitString, METH_VARARGS,
+      "emitString(section, min_order) -> string\n\n"
+      "Return a string with the marked options for section, with at least min_order order dependency.\n"
+      "@type section: integer\n"
+      "@param section: section id\n"
+      "@type min_order: float\n"
+      "@param min_order: minumum order dependency\n"
+      "@return: string containing emitted postscript" },
+
+    { "emit",
+      (PyCFunction) PPD_emit, METH_VARARGS,
+      "emit(file, section) -> None\n\n"
+      "Output marked options for section to a file.\n"
+      "@type file: file\n"
+      "@param file: file object\n"
+      "@type section: integer\n"
+      "@param section: section id" },
+
+    { "emitAfterOrder",
+      (PyCFunction) PPD_emitAfterOrder, METH_VARARGS,
+      "emitAfterOrder(file, section, limit, min_order) -> None\n\n"
+      "Output marked options for section to a file.\n"
+      "@type file: file\n"
+      "@param file: file object\n"
+      "@type section: integer\n"
+      "@param section: section id\n"
+      "@type limit: integer\n"
+      "@param limit: non-zero to use min_order\n"
+      "@type min_order: float\n"
+      "@param min_order: minumum order dependency" },
+
+    { "emitFd",
+      (PyCFunction) PPD_emitFd, METH_VARARGS,
+      "emitFd(fd, section) -> None\n\n"
+      "Output marked options for section to a file descriptor.\n"
+      "@type fd: integer\n"
+      "@param fd: file descriptor\n"
+      "@type section: integer\n"
+      "@param section: section id" },
+
+    { "emitJCL",
+      (PyCFunction) PPD_emitJCL, METH_VARARGS,
+      "emitJCL(file, job_id, user, title) -> None\n\n"
+      "Emit code for JCL options to a file.\n"
+      "@type file: file object\n"
+      "@param file: file\n"
+      "@type job_id: integer\n"
+      "@param job_id: job id\n"
+      "@type user: string\n"
+      "@param user: user name on job\n"
+      "@type title: string\n"
+      "@param title: title of job" },
+
+    { "emitJCLEnd",
+      (PyCFunction) PPD_emitJCLEnd, METH_VARARGS,
+      "emitJCLEnd(file) -> None\n\n"
+      "Emit JCLEnd code to a file.\n"
+      "@type file: file object\n"
+      "@param file: file" },
+
     { "writeFd",
       (PyCFunction) PPD_writeFd, METH_VARARGS,
       "writeFd(fd) -> None\n\n"
@@ -761,6 +1052,17 @@ Option_dealloc (Option *self)
   self->ob_type->tp_free ((PyObject *) self);
 }
 
+static PyObject *
+Option_repr (Option *self)
+{
+  ppd_option_t *option = self->option;
+  if (!option)
+    return PyString_FromString ("<cups.Option>");
+
+  return PyString_FromFormat ("<cups.Option %s=%s>",
+			      option->keyword, option->defchoice);
+}
+
 ////////////
 // Option // ATTRIBUTES
 ////////////
@@ -833,12 +1135,20 @@ Option_getChoices (Option *self, void *closure)
        i < self->option->num_choices;
        i++, choice++) {
     PyObject *choice_dict = PyDict_New ();
-    PyDict_SetItemString (choice_dict, "choice",
-			  make_PyUnicode_from_ppd_string (self->ppd,
-							  choice->choice));
-    PyDict_SetItemString (choice_dict, "text",
-			  make_PyUnicode_from_ppd_string (self->ppd,
-							  choice->text));
+    PyObject *u;
+
+    u = make_PyUnicode_from_ppd_string (self->ppd, choice->choice);
+    PyDict_SetItemString (choice_dict, "choice", u);
+    Py_DECREF (u);
+
+    u = make_PyUnicode_from_ppd_string (self->ppd, choice->text);
+    PyDict_SetItemString (choice_dict, "text", u);
+    Py_DECREF (u);
+
+    u = PyBool_FromLong (choice->marked);
+    PyDict_SetItemString (choice_dict, "marked", u);
+    Py_DECREF (u);
+
     PyList_Append (choices, choice_dict);
     if (!strcmp (choice->choice, self->option->defchoice))
       defchoice_seen = true;
@@ -849,14 +1159,18 @@ Option_getChoices (Option *self, void *closure)
     // This really happens.
     const char *defchoice = self->option->defchoice;
     PyObject *choice_dict = PyDict_New ();
-    PyDict_SetItemString (choice_dict, "choice",
-			  make_PyUnicode_from_ppd_string (self->ppd,
-							    defchoice));
-      PyDict_SetItemString (choice_dict, "text",
-			    make_PyUnicode_from_ppd_string (self->ppd,
-							    defchoice));
-      PyList_Append (choices, choice_dict);
-    }
+    PyObject *u;
+
+    u = make_PyUnicode_from_ppd_string (self->ppd, defchoice);
+    PyDict_SetItemString (choice_dict, "choice", u);
+    Py_DECREF (u);
+
+    u = make_PyUnicode_from_ppd_string (self->ppd, defchoice);
+    PyDict_SetItemString (choice_dict, "text", u);
+    Py_DECREF (u);
+
+    PyList_Append (choices, choice_dict);
+  }
 
   return choices;
 }
@@ -907,7 +1221,7 @@ PyTypeObject cups_OptionType =
     0,                         /*tp_getattr*/
     0,                         /*tp_setattr*/
     0,                         /*tp_compare*/
-    0,                         /*tp_repr*/
+    (reprfunc)Option_repr,     /*tp_repr*/
     0,                         /*tp_as_number*/
     0,                         /*tp_as_sequence*/
     0,                         /*tp_as_mapping*/
@@ -980,6 +1294,16 @@ Group_dealloc (Group *self)
   self->ob_type->tp_free ((PyObject *) self);
 }
 
+static PyObject *
+Group_repr (Group *self)
+{
+  ppd_group_t *group = self->group;
+  if (!group)
+    return PyString_FromString ("<cups.Group>");
+
+  return PyString_FromFormat ("<cups.Group %s>", group->name);
+}
+
 ///////////
 // Group // ATTRIBUTES
 ///////////
@@ -1100,7 +1424,7 @@ PyTypeObject cups_GroupType =
     0,                         /*tp_getattr*/
     0,                         /*tp_setattr*/
     0,                         /*tp_compare*/
-    0,                         /*tp_repr*/
+    (reprfunc)Group_repr,      /*tp_repr*/
     0,                         /*tp_as_number*/
     0,                         /*tp_as_sequence*/
     0,                         /*tp_as_mapping*/
@@ -1317,6 +1641,19 @@ Attribute_dealloc (Attribute *self)
   self->ob_type->tp_free ((PyObject *) self);
 }
 
+static PyObject *
+Attribute_repr (Attribute *self)
+{
+  ppd_attr_t *attribute = self->attribute;
+  if (!attribute)
+    return PyString_FromString ("<cups.Attribute>");
+
+  return PyString_FromFormat ("<cups.Attribute *%s%s%s>",
+			      attribute->name,
+			      attribute->spec[0] != '\0' ? " ": "",
+			      attribute->spec);
+}
+
 ///////////////
 // Attribute // ATTRIBUTES
 ///////////////
@@ -1398,7 +1735,7 @@ PyTypeObject cups_AttributeType =
     0,                         /*tp_getattr*/
     0,                         /*tp_setattr*/
     0,                         /*tp_compare*/
-    0,                         /*tp_repr*/
+    (reprfunc)Attribute_repr,  /*tp_repr*/
     0,                         /*tp_as_number*/
     0,                         /*tp_as_sequence*/
     0,                         /*tp_as_mapping*/
